"""
=========
highiq.io
=========

.. currentmodule:: highiq.io

This module contains the I/O methods for loading data into and saving data from HighIQ analyses.

.. autosummary::
    :toctree: generated/

    load_arm_netcdf

"""
from .arm_data import load_arm_netcdf


import numpy as np
import cupy as cp
import xarray as xr


def _gpu_calc_power(psd, dV, block_size=200, normed=True):
    shp = psd.shape
    times = shp[0]
    power = np.zeros((shp[0], shp[1]))
    if len(shp) == 3:
        for k in range(0, times, block_size):
            the_max = min([k + block_size, times])
            gpu_array = cp.array(psd[k:the_max, :, :])
            if normed:
                gpu_array = 10 ** (gpu_array / 10. * dV)
            gpu_array = cp.sum(gpu_array, axis=2)
            power[k:the_max] = gpu_array.get()
    else:
        gpu_array = cp.array(psd)
        if normed:
            gpu_array = 10 ** (gpu_array / 10. * dV)
        gpu_array = cp.sum(gpu_array, axis=1)
        power = gpu_array.get()

    return power


def _gpu_calc_velocity(psd, power, vel_bins, dV, block_size=100):
    shp = psd.shape
    times = shp[0]
    velocity = np.zeros((shp[0], shp[1]))

    for k in range(0, times, block_size):
        the_max = min([k + block_size, times])
        gpu_array = cp.array(psd[k:the_max, :, :], dtype=float)
        power_array = cp.array(power[k:the_max, :], dtype=float)
        vel_bins_tiled = cp.tile(vel_bins, (the_max - k, shp[1], 1))
        gpu_array = 10 ** (gpu_array / 10. * dV)
        gpu_array = 1 / power_array * cp.sum(gpu_array * vel_bins_tiled, axis=2)
        velocity[k:the_max, :] = gpu_array.get()
    return velocity


def _gpu_calc_velocity_dumb(psd, vel_bins, block_size=500):
    shp = psd.shape
    times = shp[0]
    velocity = np.zeros((shp[0], shp[1]))
    dV = np.diff(vel_bins)[0]
    vel_min = vel_bins.min()
    for k in range(0, times, block_size):
        the_max = min([k + block_size, times])
        gpu_array = cp.array(psd[k:the_max, :, :])
        gpu_array = cp.argmax(gpu_array, axis=2)
        gpu_array = vel_min + gpu_array * dV
        velocity[k:the_max, :] = gpu_array.get()
    return velocity


def _gpu_calc_spectral_width(psd, power, vel_bins, velocity, dV, block_size=100):
    shp = psd.shape
    times = shp[0]
    specwidth = np.zeros((shp[0], shp[1]))
    for k in range(0, times, block_size):
        the_max = min([k+block_size, times])
        gpu_array = cp.array(psd[k:the_max, :, :], dtype=float)
        power_array = cp.array(power[k:the_max, :], dtype=float)
        velocity_array = cp.array(velocity[k:the_max, :])
        velocity_array = cp.transpose(cp.tile(velocity_array, (shp[2], 1, 1)), [1, 2, 0])
        vel_bins_tiled = cp.tile(vel_bins, (the_max-k, shp[1], 1))
        gpu_array = 10**(gpu_array/10.*dV)
        gpu_array = cp.sqrt(1 / power_array *
                            cp.sum((vel_bins_tiled - velocity_array)**2 * gpu_array, axis=2))
        specwidth[k:the_max, :] = gpu_array.get()
    return specwidth


def _gpu_snr(power, noise, block_size=200):
    shp = power.shape
    power_array = cp.array(power)
    gpu_noise = cp.array(noise)
    power_array = 10*cp.log10(power_array/gpu_noise)
    snr = power_array.get()
    return snr


def _gpu_calc_skewness(psd, power, vel_bins, velocity, spec_width, dV, block_size=100):
    shp = psd.shape
    times = shp[0]
    skewness = np.zeros((shp[0], shp[1]))
    for k in range(0, times, block_size):
        the_max = min([k+block_size, times])
        gpu_array = cp.array(psd[k:the_max, :, :], dtype=float)
        power_array = cp.array(power[k:the_max, :], dtype=float)
        spec_width_array = cp.array(spec_width[k:the_max, :], dtype=float)
        power_array *= spec_width_array**3
        del spec_width_array
        velocity_array = cp.array(velocity[k:the_max, :])
        velocity_array = cp.transpose(cp.tile(velocity_array, (shp[2], 1, 1)), [1, 2, 0])
        vel_bins_tiled = cp.tile(vel_bins, (the_max-k, shp[1], 1))
        gpu_array = 10**(gpu_array/10.*dV)
        gpu_array = 1/power_array*cp.sum((vel_bins_tiled - velocity_array)**3 * gpu_array, axis=2)
        skewness[k:the_max, :] = gpu_array.get()
    return skewness


def _gpu_calc_kurtosis(psd, power, vel_bins, velocity, spec_width, dV, block_size=100):
    shp = psd.shape
    times = shp[0]
    kurtosis = np.zeros((shp[0], shp[1]))
    for k in range(0, times, block_size):
        the_max = min([k+block_size, times])
        gpu_array = cp.array(psd[k:the_max, :, :], dtype=float)
        power_array = cp.array(power[k:the_max, :], dtype=float)
        spec_width_array = cp.array(spec_width[k:the_max, :], dtype=float)
        power_array *= spec_width_array**4
        velocity_array = cp.array(velocity[k:the_max, :])
        velocity_array = cp.transpose(cp.tile(velocity_array, (shp[2], 1, 1)), [1, 2, 0])
        vel_bins_tiled = cp.tile(vel_bins, (the_max-k, shp[1], 1))
        gpu_array = 10**(gpu_array/10.*dV)
        gpu_array = 1/power_array*cp.sum((vel_bins_tiled - velocity_array)**4 * gpu_array, axis=2)
        kurtosis[k:the_max, :] = gpu_array.get()
    return kurtosis


def get_lidar_moments(spectra, snr_thresh=0, block_size_ratio=1.0, which_moments=None):
    """
    This function will retrieve the lidar moments of the Doppler spectra.

    Parameters
    ----------
    spectra: ACT Dataset
        The dataset containing the processed Doppler spectral density functions.
    snr_thresh: float
        The minimum signal to noise ratio to use as an initial mask of noise.
    block_size_ratio: float
        This value is used to determine how much data the GPU will process in one loop. If your
        GPU has more memory, you may be able to optimize processing by raising this number. In
        addition, if you encounter out of memory errors, try lowering this number, ensuring that
        it is a positive floating point number.
    which_moments: list or None
        This tells HighIQ which moments should be processed. If this list is None, then the
        signal to noise ratio, doppler velocity, spectral width, skewness,
        and kurtosis will be calculated.

    Returns
    -------
    spectra: ACT Dataset
        The database with the Doppler lidar moments.
    """
    if which_moments is None:
        which_moments = ['snr', 'doppler_velocity', 'spectral_width',
                         'skewness', 'kurtosis']
    else:
        which_moments = [x.lower() for x in which_moments]

    if not block_size_ratio > 0:
        raise ValueError("block_size_ratio must be a positive floating point number!")

    dV = np.diff(spectra['vel_bins'])[0]
    linear_psd = spectra['power_spectral_density_interp']
    linear_psd_0filled = linear_psd.fillna(0)
    power = _gpu_calc_power(
        linear_psd_0filled, dV, block_size=round(200 * block_size_ratio))
    if ('doppler_velocity' in which_moments or 'spectral_wifth' in which_moments or
        'skewness' in which_moments or 'kurtosis' in which_moments):
        velocity = _gpu_calc_velocity(
            linear_psd_0filled, power,
            spectra['vel_bin_interp'].values, dV, block_size=round(100*block_size_ratio))

    spectra['noise'] = spectra['power_bkg'].sum(axis=2)
    if 'snr' in which_moments:
        power_with_noise = _gpu_calc_power(
            spectra['power'].values, dV,
            normed=False, block_size=round(200 * block_size_ratio))
        power_with_noise = xr.DataArray(power_with_noise, dims=('time', 'range'))
        spectra['snr'] = xr.DataArray(
            _gpu_snr(power_with_noise, spectra['noise'].values),
            dims=('time', 'range'))
        spectra['snr'].attrs['long_name'] = "Signal to Noise Ratio"
        spectra['snr'].attrs['units'] = "dB"
        spectra['intensity'] = spectra['snr'] + 1.
        spectra['intensity'].attrs['long_name'] = "Intensity (SNR + 1)"
        spectra['intensity'].attrs['units'] = "dB"
        spectra['intensity'] = \
            spectra['intensity'].where(spectra.snr > snr_thresh)
        spectra.attrs['snr_mask'] = "%f dB" % snr_thresh

    if 'doppler_velocity' in which_moments:
        velocity_dumb = _gpu_calc_velocity_dumb(
            linear_psd_0filled, spectra['vel_bin_interp'].values,
            block_size=round(500*block_size_ratio))
        spectra['doppler_velocity_max_peak'] = xr.DataArray(
            velocity_dumb, dims=('time', 'range'))
        spectra['doppler_velocity_max_peak'].attrs['long_name'] = \
            "Doppler velocity derived using location of highest " \
            "peak in spectra."
        spectra['doppler_velocity_max_peak'].attrs["units"] = "m s-1"
        spectra['doppler_velocity'] = xr.DataArray(
            velocity, dims=('time', 'range'))
        spectra['doppler_velocity'].attrs['long_name'] = \
            "Doppler velocity using first moment"
        spectra['doppler_velocity'].attrs['units'] = "m s-1"
        spectra['doppler_velocity_max_peak'] = \
            spectra['doppler_velocity_max_peak'].where(spectra.snr > snr_thresh)
        spectra['doppler_velocity'] = spectra['doppler_velocity'].where(
            spectra.snr > snr_thresh)

    if 'spectral_width' in which_moments or 'kurtosis' in which_moments or 'skewness' in which_moments:
        spectral_width = _gpu_calc_spectral_width(
            linear_psd, power, spectra['vel_bin_interp'].values,
            velocity, dV, block_size=round(100 * block_size_ratio))

    if 'spectral_width' in which_moments:
        spectra['spectral_width'] = xr.DataArray(
            spectral_width, dims=('time', 'range'))
        spectra['spectral_width'].attrs["long_name"] = "Spectral width"
        spectra['spectral_width'].attrs["units"] = "m s-1"
        if 'snr' in which_moments:
            spectra['spectral_width'] = spectra['spectral_width'].where(spectra.snr > snr_thresh)

    if 'skewness' in which_moments:
        skewness = _gpu_calc_skewness(
            linear_psd, power, spectra['vel_bin_interp'].values, velocity, spectral_width, dV,
            block_size=round(100 * block_size_ratio))
        spectra['skewness'] = xr.DataArray(skewness, dims=('time', 'range'))
        if 'snr' in which_moments:
            spectra['skewness'] = spectra['skewness'].where(spectra.snr > snr_thresh)
        spectra['skewness'].attrs["long_name"] = "Skewness"
        spectra['skewness'].attrs["units"] = "m^3 s^-3"

    if 'kurtosis' in which_moments:
        kurtosis = _gpu_calc_kurtosis(
            linear_psd, power, spectra['vel_bin_interp'].values, velocity, spectral_width, dV,
            block_size=round(100 * block_size_ratio))
        spectra['kurtosis'] = xr.DataArray(kurtosis, dims=('time', 'range'))
        if 'snr' in which_moments:
            spectra['kurtosis'] = spectra['kurtosis'].where(spectra.snr > snr_thresh)
        spectra['skewness'].attrs["long_name"] = "Kurtosis"
        spectra['skewness'].attrs["units"] = "m^4 s^-4"

    spectra['range'].attrs['long_name'] = "Range"
    spectra['range'].attrs['units'] = 'm'
    spectra['vel_bins'].attrs['long_name'] = "Doppler velocity"
    spectra['vel_bins'].attrs['units'] = 'm s-1'
    return spectra


"""
===========
highiq.calc
===========

.. currentmodule:: highiq.calc

This module contains the methods that do the core calculations of HighIQ.

.. autosummary::
    :toctree: generated/

    get_lidar_moments
    get_psd
    calc_num_peaks
    welchs_method

"""
from .moments import get_lidar_moments
from .spectra import get_psd, calc_num_peaks, welchs_method



import numpy as np
import cupy as cp
import xarray as xr

from scipy.signal import hann, find_peaks
from pint import UnitRegistry
from .gpu_methods import _fast_expand, _gpu_moving_average


def welchs_method(complex_coeff, fs=50e6, nfft=32, window_skip=16, num_per_block=200):
    """
    This technique is an implementation of Welch's method for processing a Doppler spectral
    density function from complex autocorrelation function data. Welch's method calculates
    the Doppler spectral density function by calculating the FFT over (potentially overlapping)
    windows and then taking the average of the magnitudes of each FFT to create the Doppler spectra.

    Parameters
    ----------
    complex_coeff: complex nD array
        An n-D array of complex floats representing the value of the autocorrelation
        function. The first dimension is the number of entries in time, and the second is the number
        of entries in the range dimension.
    fs:
        The pulse repetition frequency of the radar.
    nfft:
        The number of points to include in the FFT for calculating spectra. This must
        be an even number.
    window_skip:
        The number of points to go forward for each window. This cannot exceed the
        size of the FFT.
    num_per_block:
        The number of time periods to send to the GPU at one time. This is implemented
        because the GPU can only handle a limited number of data points at one time.

    Returns
    -------
    freq: 1D array
        The frequencies corresponding to each value in the spectra.
    power: nD array
        The power spectral densities for each frequency.
    """

    if nfft % 2 == 1:
        raise ValueError("The number of points in the FFT must be even!")

    if window_skip > nfft:
        raise ValueError("window_skip cannot be greater than nfft!")

    times = complex_coeff.shape[0]
    num_points = complex_coeff.shape[-1]
    window = hann(nfft)
    if len(complex_coeff.shape) == 3:
        my_fft = np.zeros((complex_coeff.shape[0], complex_coeff.shape[1], nfft))
        for k in range(0, times, num_per_block):
            j = 0
            the_max = min([k + num_per_block, times])
            gpu_complex = cp.array(complex_coeff[k:the_max, :, :])
            windowt = cp.tile(window, (gpu_complex.shape[0], gpu_complex.shape[1], 1))
            temp_fft = cp.zeros((gpu_complex.shape[0], gpu_complex.shape[1], nfft))
            for i in range(0, num_points, window_skip):
                if i + nfft > num_points:
                    start = num_points - nfft
                    temp_fft += cp.square(
                        cp.abs(cp.fft.fft(gpu_complex[:, :, start:] * windowt, axis=-1)))
                else:
                    temp_fft += cp.square(
                        cp.abs(cp.fft.fft(gpu_complex[:, :, i:i + nfft] * windowt, axis=-1)))
                j += 1
            temp_fft = temp_fft / j
            my_fft[k:the_max] = temp_fft.get()
    elif len(complex_coeff.shape) == 2:
        j = 0
        gpu_complex = cp.array(complex_coeff)
        windowt = cp.tile(window, ((gpu_complex.shape[0], 1)))
        temp_fft = cp.zeros((gpu_complex.shape[0], nfft))
        for i in range(0, num_points, window_skip):
            if i + nfft > num_points:
                start = num_points - nfft
                temp_fft += cp.square(
                    cp.abs(cp.fft.fft(gpu_complex[:, start:] * windowt, axis=-1)))
            else:
                temp_fft += cp.square(
                    cp.abs(cp.fft.fft(gpu_complex[:, i:i + nfft] * windowt, axis=-1)))
            j += 1
        temp_fft = temp_fft / j
        my_fft = temp_fft.get()
    power = my_fft
    freq = np.fft.fftfreq(nfft) * fs

    return freq, np.array(power)


def get_psd(spectra, gate_resolution=30., wavelength=None, fs=None, nfft=32,
            acf_name='acf', acf_bkg_name='acf_bkg', block_size_ratio=1.0):
    """
    This function will get the power spectral density from the autocorrelation function.

    Parameters
    ----------
    spectra: ACT Dataset
        The dataset containing the autocorrelation function.
    gate_resolution: float
        The gate resolution to derive the spectra for.
    wavelength: float or None
        The wavelength (in m) of the radar. If None, HighIQ will attempt to load the
        wavelength from the wavelength attribute of the spectra dataset.
    fs: float or None
        The pulse repetition frequency of the radar in Hz. If None,
        HighIQ will try and extract this information from the
        ACT dataset. This will require that the dataset
        have an attribute called sample_rate that contains a string
        with the magnitude and units of the sample rate.
    nfft: int
        The number of points to include in the FFT.
    acf_name: str
        The name of the autocorrelation function field.
    acf_bkg_name: str
        The name of the autocorrelation function of the background.
    block_size_ratio: float
        Increase this value to use more GPU memory for processing. Doing this can
        poentially optimize processing.

    Returns
    -------
    spectra: ACT Dataset
        The dataset containing the power spectral density functions.
    """

    Q_ = UnitRegistry().Quantity
    if fs is None:
        if not "sample_rate" in spectra.attrs:
            raise KeyError("If sample frequency is not specified, then ACT Dataset must contain a sample_rate " +
                           "attribute!")

        fs_pint = Q_(spectra.attrs["sample_rate"])
        fs_pint = fs_pint.to("Hz")
        fs = fs_pint.magnitude


    if wavelength is None:
        if not "wavelength" in spectra.attrs:
            raise KeyError("If wavelength is not specified, then the dataset must contain a wavelength attribute!")
        fs_pint = Q_(spectra.attrs["wavelength"])
        fs_pint = fs_pint.to("m")
        wavelength = fs_pint.magnitude


    num_gates = int(gate_resolution / 3)
    complex_coeff = spectra[acf_name].sel(complex=1).values +\
                    spectra[acf_name].sel(complex=2).values * 1j
    complex_coeff = np.reshape(
        complex_coeff, (complex_coeff.shape[0],
                        int(complex_coeff.shape[1] / num_gates),
                        int(complex_coeff.shape[2] * num_gates)))
    freq, power = welchs_method(
        complex_coeff, fs=fs, nfft=nfft, num_per_block=int(block_size_ratio*200))
    attrs_dict = {'long_name': 'Range', 'units': 'm'}
    spectra['range'] = xr.DataArray(
        gate_resolution * np.arange(int(complex_coeff.shape[1])),
        dims=('range'), attrs=attrs_dict)

    spectra.attrs['nyquist_velocity'] = "%f m s-1" % (wavelength / (4 * 1 / fs))
    spectra['freq_bins'] = xr.DataArray(freq, dims=['freq'])
    spectra['freq_bins'].attrs["long_name"] = "Doppler spectra bins in frequency units"
    spectra['freq_bins'].attrs["units"] = "Hz"
    vel_bins = spectra['freq_bins'] * (wavelength / 2)
    inds_sorted = np.argsort(vel_bins.values)
    vel_bins = vel_bins[inds_sorted]
    attrs_dict = {'long_name': 'Doppler velocity', 'units': 'm s-1'}
    spectra['vel_bins'] = xr.DataArray(vel_bins, dims=('vel_bins'), attrs=attrs_dict)
    spectra['freq_bins'] = spectra['freq_bins'][inds_sorted]
    spectra['power'] = xr.DataArray(
        power[:, :, inds_sorted], dims=(('time', 'range', 'vel_bins')))

    complex_coeff = (spectra[acf_bkg_name].sel(complex=1).values +
                     spectra[acf_bkg_name].sel(complex=2).values * 1j)
    complex_coeff = np.reshape(
        complex_coeff, (complex_coeff.shape[0],
                        int(complex_coeff.shape[1] / num_gates),
                        int(complex_coeff.shape[2] * num_gates)))
    freq, power = welchs_method(
        complex_coeff, fs=50e6, nfft=32, num_per_block=int(200*block_size_ratio))
    spectra['power_bkg'] = xr.DataArray(
        power[:, :, inds_sorted], dims=(('time', 'range', 'vel_bins')))

    # Subtract background noise
    spectra['power_spectral_density'] = spectra['power'] - spectra['power_bkg']
    spectra['power_spectral_density'] = spectra['power_spectral_density'].where(
        spectra['power_spectral_density'] > 0, 0)
    spectra['power_spectral_density'].attrs["long_name"] = "Power spectral density"
    tot_power = np.sum(spectra['power_spectral_density'].values, axis=2)
    dV = np.diff(spectra['vel_bins'])
    power_tiled = np.stack(
        [tot_power for x in range(spectra['power'].values.shape[2])], axis=2)
    spectra['power_spectra_normed'] = spectra['power_spectral_density'] / power_tiled / dV[1] * 100
    spectra['power_spectra_normed'].attrs["long_name"] = "p.d.f. of power spectra"
    spectra['power_spectra_normed'].attrs["units"] = "%"
    spectra['power_spectral_density'] = 10 * np.log10(spectra['power_spectral_density']) / dV[1]
    spectra['power_spectral_density'].attrs["units"] = 's dB-1 m-1 '

    # Smooth out power spectra
    interpolated_bins = np.linspace(
        spectra['vel_bins'].values[0], spectra['vel_bins'].values[-1], 256)
    spectra['vel_bin_interp'] = xr.DataArray(interpolated_bins, dims=('vel_bin_interp'))
    my_array = _gpu_moving_average(
        _fast_expand(spectra['power_spectral_density'].values, 8))
    spectra['power_spectral_density_interp'] = xr.DataArray(
        my_array, dims=('time', 'range', 'vel_bin_interp'))
    spectra['power_spectral_density_interp'].attrs['long_name'] = "Power spectral density"
    spectra['power_spectral_density_interp'].attrs["units"] = "s dB-1 m-1"
    my_array = _gpu_moving_average(
        _fast_expand(spectra['power_spectra_normed'].values, 8))
    spectra['power_spectra_normed_interp'] = xr.DataArray(
        my_array, dims=('time', 'range', 'vel_bin_interp'),)
    spectra['power_spectra_normed_interp'].attrs['long_name'] = "p.d.f of power spectra"
    spectra['power_spectra_normed_interp'].attrs["units"] = "%"

    spectra['range'].attrs['long_name'] = "Range"
    spectra['range'].attrs['units'] = 'm'
    spectra['vel_bins'].attrs['long_name'] = "Doppler velocity"
    spectra['vel_bins'].attrs['units'] = 'm s-1'
    spectra['vel_bin_interp'].attrs['long_name'] = "Doppler velocity"
    spectra['vel_bin_interp'].attrs["units"] = "m s-1"
    return spectra


def calc_num_peaks(my_spectra, **kwargs):
    """

    This function will calculate the number of peaks in the spectra.

    Parameters
    ----------
    my_spectra: ACT Dataset
        The dataset to calculate the number of peaks for.
    kwargs:
        Additional keyword arguments are passed into :func:`scipy.signal.find_peaks`.
        The default minimum height and width of the peak are set to 3 and 8 points
        respectively.

    Returns
    -------
    my_spectra: ACT Dataset
        The dataset with an 'npeaks' variable included that shows the number of peaks.
    """
    spectra = my_spectra['power_spectra_normed_interp']
    my_array = spectra.fillna(0).values
    shp = my_array.shape
    num_peaks = np.zeros((shp[0], shp[1]))

    if not 'height' in kwargs.keys():
        height = 3
    else:
        height = kwargs.pop('height')

    if not 'width' in kwargs.keys():
        width = 8
    else:
        width = kwargs.pop('height')

    for i in range(shp[0]):
        for j in range(shp[1]):
            num_peaks[i, j] = len(
                find_peaks(my_array[i,j], height=height, width=width, **kwargs)[0])
    my_spectra['npeaks'] = xr.DataArray(num_peaks, dims=('time', 'range'))
    my_spectra['npeaks'].attrs['long_name'] = "Number of peaks in Doppler spectra"
    my_spectra['npeaks'].attrs['units'] = "1"

    return my_spectra


"""
This module contains test settings and files that are useful for
unit testing HighIQ.

You can test the processing from a sample ACF file by loading
:func:`highiq.testing.TEST_FILE`.
"""

import os

DATA_PATH = os.path.join(os.path.dirname(__file__), 'data')
TEST_FILE = os.path.join(DATA_PATH, 'testsig.a1.20180801.000003.nc')


import act


def load_arm_netcdf(arm_file, **kwargs):
    """

    This loads netCDF data that are in the Atmospheric Radiation Measurement standard netCDF format.
    This is a wrapper around :func:`act.io.armfiles.read_netcdf`

    Parameters
    ----------
    arm_file: str
        The path to the dataset to load.

    Additional keyword arguments are passed into :func:`act.io.armfiles.read_netcdf`

    Returns
    -------
    ds: ACT Dataset
        Returns the ACT dataset (xarray dataset) that contains the autocorrelation functions.
    """
    return act.io.armfiles.read_netcdf(arm_file, **kwargs)



import numpy as np
import cupy as cp

def _fast_expand(complex_array, factor, num_per_block=200):
    shp = complex_array.shape
    times = shp[0]
    expanded_array = np.zeros((shp[0], shp[1], shp[2] * factor))
    for k in range(0, times, num_per_block):
        the_max = min([k + num_per_block, times])
        weights = cp.tile(np.arange(0, factor) / factor, (the_max - k, shp[1], 1))
        for l in range(shp[2]):
            gpu_array = cp.array(complex_array[k:the_max, :, l])
            if l < shp[2] - 1:
                gpu_array2 = cp.array(complex_array[k:the_max, :, l + 1])
                diff_array = gpu_array2 - gpu_array
            else:
                diff_array = cp.zeros((the_max - k, shp[1]))

            rep_array = cp.transpose(
                cp.tile(gpu_array, (factor, 1, 1)), [1, 2, 0])
            diff_array = cp.transpose(
                cp.tile(diff_array, (factor, 1, 1)), [1, 2, 0])
            temp_array = diff_array * weights + rep_array
            expanded_array[k:the_max, :, factor * l:factor * (l + 1)] = temp_array.get()
    return expanded_array


def _gpu_moving_average(arr, window=8, num_per_block=200):
    shp = arr.shape
    times = shp[0]
    for k in range(0, times, num_per_block):
        the_max = min([k + num_per_block, times])
        gpu_arr = cp.array(arr[k:the_max, :, :])
        gpu_arr2 = cp.zeros_like(gpu_arr)
        for i in range(shp[2]):
            the_min2 = max([0, i - int(window / 2)])
            the_max2 = min([i + int(window / 2), shp[2]])
            gpu_arr2[:, :, i] = cp.mean(gpu_arr[:, :, the_min2:the_max2], axis=2)
        arr[k:the_max, ::] = gpu_arr2.get()
        del gpu_arr, gpu_arr2
    return arr


# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = 'HighIQ'
copyright = '2019, Bobby Jackson'
author = 'Bobby Jackson'

# The full version, including alpha/beta/rc tags
release = '1.0'

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.githubpages',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'IPython.sphinxext.ipython_directive',
    'IPython.sphinxext.ipython_console_highlighting',
    'matplotlib.sphinxext.plot_directive',
    'numpydoc',
    'sphinx_copybutton',
    'sphinx_gallery.gen_gallery',
]

# -- General configuration ---------------------------------------------------

sphinx_gallery_conf = {
    'examples_dirs': '../../examples',
    'gallery_dirs': 'source/auto_examples'
}

#Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
html_theme = 'neo_rtd_theme'
import sphinx_theme
html_theme_path = [sphinx_theme.get_html_theme_path()]

# Numpy autodoc attributes
numpydoc_show_class_members = True
autosummary_generate = True